from functools import wraps

import numpy as np


def stvariogram(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        st = args[0]
        if st.ndim == 2:
            new_args = args[1:]
            mapping = map(lambda lags: func(lags, *new_args, **kwargs), st)
            return np.fromiter(mapping, dtype=float)
        else:
            return func(*args, **kwargs)
    return wrapper


@stvariogram
def sum(lags, Vx, Vt):
    """Sum space-time model

    Separable space-time variogram model. This is the most basic model as the
    two marginal models of the space and time axis are simply summed up for
    each lag pair. Further, there are no fitting parameters.
    Please consider the notes before using this model.

    Parameters
    ----------
    lags : tuple
        Tuple of the space (x) and time (t) lag given as tuple: (x, t) which
        will be used to calculate the dependent semivariance.
    Vx : skgstat.Variogram.fitted_model
        instance of the space marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.
    Vt : skgstat.Variogram.fitted_model
        instance of the time marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.

    Returns
    -------
    gamma : float
        The semi-variance modeled for the given lags.

    Notes
    -----
    This model is implemented like:

    .. math::
        \gamma (h,t) = \gamma_x (h) + \gamma_t (t)
    
    Where :math:`\gamma_x(h)` is the spatial marginal variogram and 
    :math:`\gamma_t(t)` is the temporal marginal variogram.

    It is not a good idea to use this model in almost any case, as it assumes
    the covariance field to be isotropic in space and time direction,
    which will hardly be true. Further, it might not be strictly definite as
    shown by [7]_, [8]_, [9]_.

    References
    ----------
    .. [7] Myers, D. E., Journel, A. (1990), Variograms with Zonal
       Anisotropies and Non-Invertible Kriging Systems.
       Mathematical Geology 22, 779-785.
    .. [8] Dimitrakopoulos, R. and Lou, X. (1994), Spatiotemporal modeling:
       covariances and ordinary kriging systems, in R. Dimitrakopoulos,
       (ed.) Geostatistics for the next century, Kluwer Academic Publishers,
       Dodrecht 88-93.

    """
    h, t = lags
    return Vx(h) + Vt(t)


@stvariogram
def product(lags, Vx, Vt, Cx, Ct):
    """Product model

    Separable space-time variogram model. This model is based on the product
    of the marginal space and time models.

    Parameters
    ----------
    lags : tuple
        Tuple of the space (x) and time (t) lag given as tuple: (x, t) which
        will be used to calculate the dependent semivariance.
    Vx : skgstat.Variogram.fitted_model
        instance of the space marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.
    Vt : skgstat.Variogram.fitted_model
        instance of the time marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.
    Cx : float
        Marginal space sill.
    Ct : float
        Marignal time sill.

    Returns
    -------
    gamma : float
        The semi-variance modeled for the given lags.

    Notes
    -----
    The product sum model is implemented following [14]_:

    .. math::
        \gamma (h,t) = C_x * \gamma_t(t) + C_t * \gamma_x(h) - \gamma_x(h) * \gamma_t(t) 
    
    Where :math:`\gamma_x(h)` is the spatial marginal variogram and 
    :math:`\gamma_t(t)` is the temporal marginal variogram.

    References
    ----------
    .. [14] De Cesare, L., Myers, D., and Pose, D. (201b), FORTRAN 77 programs
       for space-time modeling, Computers & Geoscience 28, 205-212.

    """
    h, t = lags
    return Cx * Vt(t) + Ct * Vx(h) - Vx(h) * Vt(t)


@stvariogram
def product_sum(lags, Vx, Vt, k1, k2, k3, Cx, Ct):
    """Product-Sum space-time model

    Separable space-time variogram model, based on a combination of 'sum' and
    'product' models. Both base models are based on separated marginal
    variograms for the space and time axis.

    Parameters
    ----------
    lags : tuple
        Tuple of the space (x) and time (t) lag given as tuple: (x, t) which
        will be used to calculate the dependent semivariance.
    Vx : skgstat.Variogram.fitted_model
        instance of the space marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.
    Vt : skgstat.Variogram.fitted_model
        instance of the time marginal variogram with a fitted theoretical
        model sufficiently describing the marginal. If this model does not fit
        the experimental variogram, the space-time model fit will be poor as
        well.
    k1 : float
        Fitting parameter. k1 has to be positive or zero and may not be larger
        than all marginal sill values.
    k2 : float
        Fitting paramter. k2 has to be positive or zero and may not be larger
        than all marginal sill values.
    k3 : float
        Fitting parameter. k3 has to be positive and may not be larger than
        all marginal sill values.
    Cx : float
        Marginal space sill.
    Ct : float
        Marignal time sill.

    Returns
    -------
    gamma : float
        The semi-variance modeled for the given lags.

    Notes
    -----
    This model implements the product-sum model as suggested by
    De Cesare et. al [15]_, [16]_:

    .. math::
        \\gamma_{ST}(h_s, h_t) = [k_1C_T(0) + k_2]*\\gamma_S(h_s) +
        [k_1C_s(0) + k_3]\\gamma_T(h_t) - k_1\\gamma_s(h_s) x \\gamma_T(h_t)

    References
    ----------
    .. [15] De Cesare, L., Myers, D. and Posa, D. (2001a), Product-sum
       covariance for space-time mdeling, Environmetrics 12, 11-23.
    .. [16] De Cesare, L., Myers, D., and Pose, D. (201b), FORTRAN 77 programs
       for space-time modeling, Computers & Geoscience 28, 205-212.

    """
    h, t = lags
    return (k2 + k1*Ct)*Vx(h) + (k3 + k1*Cx) * Vt(t) - k1 * Vx(h) * Vt(t)



"""
Estimate uncertainties propagated through the Variogram
using a MonteCarlo approach
"""
from typing import Union, List
from skgstat import Variogram
import numpy as np
from tqdm import tqdm
from joblib import Parallel, delayed


def propagate(
    variogram: Variogram = None,
    source: Union[str, List[str]] = 'values',
    sigma: Union[float, List[float]] = 5,
    evalf: Union[str, List[str]] = 'experimental',
    verbose: bool = False,
    use_bounds: bool = False,
    **kwargs
):
    """
    Uncertainty propagation for the variogram.
    For a given :class:`Variogram <skgstat.Variogram>`
    instance a source of error and scale of error
    distribution can be specified. The function will
    propagate the uncertainty into different parts of
    the :class:`Variogram <skgstat.Variogram>` and
    return the confidence intervals or error bounds.

    Parameters
    ----------
    variogram : skgstat.Variogram
        The base variogram. The variogram parameters will
        be used as fixed arguments for the Monte Carlo
        simulation.
    source : list
        Source of uncertainty. This has to be an attribute
        of :class:`Variogram <skgstat.Variogram>`. Right
        now only ``'values'`` is really supported, anything
        else is untested.
    sigma : list
        Standard deviation of the error distribution.
    evalf : list
        Evaluation function. This specifies, which part of
        the :class:`Variogram <skgstat.Variogram>` should be
        used to be evaluated. Possible values are
        ``'experimental'`` for the experimental variogram,
        ``'model'`` for the fitted model and ``parameter'``
        for the variogram parameters
    verbose : bool
        If True, the uncertainty_framework package used under
        the hood will print a progress bar to the console.
        Defaults to False.
    use_bounds : bool
        Shortcut to set the confidence interval bounds to the
        minimum and maximum value and thus return the error
        margins over a confidence interval.

    Keyword Arguments
    -----------------
    distribution : str
        Any valid :any:`numpy.random` distribution function, that
        takes the scale as argument.
        Defaults to ``'normal'``.
    q : int
        Width (percentile) of the confidence interval. Has to be a
        number between 0 and 100. 0 will result in the minimum and
        maximum value as bounds. 100 turns both bounds into the
        median value.
        Defaults to ``10``
    num_iter : int
        Number of iterations used in the Monte Carlo simulation.
        Defaults to ``500``.
    eval_at : int
        If evalf is set to model, the theoretical model get evaluated
        at this many evenly spaced lags up to maximum lag.
        Defaults to ``100``.
    n_jobs : int
        The evaluation can be performed in parallel. This will specify
        how many processes may be spawned in parallel. None will spwan
        only one (default).

        .. note::
            This is an untested experimental feature.

    Returns
    -------
    conf_interval : numpy.ndarray
        Confidence interval of the uncertainty propagation as
        [lower, median, upper]. If more than one evalf is given, a
        list of ndarrays will be returned.
        See notes for more details.

    Notes
    -----
    For each member of the evaluated property, the lower and upper bound
    along with the median value is retuned as ``[low, median, up]``.
    Thus the returned array has the shape ``(N, 3)``.
    N is the lengh of evaluated property, which is
    :func:`n_lags <skgstat.Variogram.n_lags` for ``'experimental'``,
    either ``3`` for ``'parameter'`` or ``4`` if
    :func:`Variogram.model = 'stable' | 'matern' <skgstat.Variogram.model>`
    and ``100`` for ``'model'`` as the model gets evaluated at
    100 evenly spaced lags up to the maximum lag class. This amount
    can be changed using the eval_at parameter.

    If more than one evalf parameter is given, the Variogram will be
    evaluated at multiple steps and each one will be returned as a
    confidence interval. Thus if ``len(evalf) == 2``, a list containing
    two confidence interval matrices will be returned.
    The order is [experimental, parameter, model].

    """
    # handle error bounds shortcut
    if use_bounds:
        kwargs['q'] = 0

    # extract the MetricSpace to speed things up a bit
    metricSpace = variogram._X

    # get the source of error
    if isinstance(source, str):
        source = [source]

    if not isinstance(sigma, (list, tuple)):
        sigma = [sigma]

    if isinstance(evalf, str):
        evalf = [evalf]

    # get the static variogram parameters
    _var_opts = variogram.describe().get('params', {})
    omit_names = [*source, 'verbose']
    args = {k: v for k, v in _var_opts.items() if k not in omit_names}

    # add back the metric space
    args['coordinates'] = metricSpace

    # build the parameter field
    num_iter = kwargs.get('num_iter', 500)
    rng = np.random.default_rng(kwargs.get('seed'))
    dist = getattr(rng, kwargs.get('distribution', 'normal'))
    param_field = []

    for it in range(num_iter):
        par = {**args}

        # add the noisy params
        for s, err in zip(source, sigma):
            obs = getattr(variogram, s)
            size = len(obs) if hasattr(obs, '__len__') else 1
            par[s] = dist(obs, err, size=size)

        # append to param field
        param_field.append(par)

    # define the eval function
    def func(par):
        vario = Variogram(**par)
        out = []
        if 'experimental' in evalf:
            out.append(vario.experimental)
        if 'parameter' in evalf:
            out.append(vario.parameters)
        if 'model' in evalf:
            x = np.linspace(0, np.max(vario.bins), kwargs.get('eval_at', 100))
            out.append(vario.fitted_model(x))
        return out

    # build the worker
    worker = Parallel(n_jobs=kwargs.get('n_jobs'))
    if verbose:
        generator = (delayed(func)(par) for par in tqdm(param_field))
    else:
        generator = (delayed(func)(par) for par in param_field)

    # run
    result = worker(generator)

    # split up conf intervals
    conf_intervals = []

    for i in range(len(evalf)):
        # unpack
        res = [result[j][i] for j in range(len(result))]

        # create the result
        ql = int(kwargs.get('q', 10) / 2)
        qu = 100 - int(kwargs.get('q', 10) / 2)
        conf_intervals.append(
            np.column_stack((
                np.percentile(res, ql, axis=0),
                np.median(res, axis=0),
                np.percentile(res, qu, axis=0)
            ))
        )

    # return
    if len(conf_intervals) == 1:
        return conf_intervals[0]
    return conf_intervals



"""
"""

import unittest

import numpy as np

from skgstat.estimators import matheron, cressie, dowd, genton
from skgstat.estimators import minmax, percentile, entropy


class TestEstimator(unittest.TestCase):
    def setUp(self):
        pass

    def test_matheron(self):
        # extract actual estimator
        e = matheron.py_func
        np.random.seed(42)

        self.assertAlmostEqual(
            e(np.random.normal(0, 1, 10000)),
            0.50342,
            places=6
        )

    def test_matheron_nan(self):
        # extract actual estimator
        e = matheron.py_func

        self.assertTrue(np.isnan(e(np.array([]))))

    def test_cressie(self):
        # extract actual estimator
        e = cressie.py_func

        np.random.seed(42)

        self.assertAlmostEqual(
            e(np.random.gamma(10, 4, 10000)),
            1686.7519,
            places=4
        )

    def test_cressie_nan(self):
        # extract actual estimator
        e = cressie.py_func

        self.assertTrue(np.isnan(e(np.array([]))))

    def test_dowd(self):
        np.random.seed(1306)
        x1 = np.random.weibull(14, 1000)
        np.random.seed(1312)
        x2 = np.random.gamma(10, 4, 100)

        # test
        self.assertAlmostEqual(dowd(x1), 2.0873, places=4)
        self.assertAlmostEqual(dowd(x2), 3170.97, places=2)

    def test_genton(self):
        # extract actual estimator
        e = genton.py_func

        np.random.seed(42)
        x1 = np.random.gamma(40, 2, 100)
        np.random.seed(42)
        x2 = np.random.gamma(30, 5, 1000)

        self.assertAlmostEqual(e(x1), 0.0089969, places=7)
        self.assertAlmostEqual(e(x2), 0.0364393, places=7)

    def test_genton_nan(self):
        # extract actual estimator
        e = genton.py_func

        # genton cannot be solved for only one element
        self.assertTrue(np.isnan(e(np.array([0.1]))))

    def test_minmax_skew(self):
        # heavily skewed gamma
        np.random.seed(1306)
        x = np.random.gamma(15, 20, 100)
        self.assertAlmostEqual(minmax(x), 1.5932, places=4)

    def test_minmax_pow(self):
        # L-stable pareto
        np.random.seed(2409)
        x = np.random.pareto(2, 10)
        self.assertAlmostEqual(minmax(x), 2.5, places=2)

    def test_percentile(self):
        np.random.seed(42)
        x = np.abs(np.random.normal(0, 1, 100000))

        self.assertAlmostEqual(percentile(x), 0.67588, places=5)
        self.assertAlmostEqual(percentile(x, 20), 0.25277, places=5)

    def test_entropy_default_bins(self):
        np.random.seed(42)
        x = np.random.normal(5, 1, 10000)

        self.assertAlmostEqual(entropy(x, bins=None), 3.0, places=2)

    def test_entropy_custom_bins(self):
        np.random.seed(123456789)
        x = np.random.gamma(10, 5, 10000)

        # custom bins
        self.assertAlmostEqual(
            entropy(x, [5, 15, 50, 51, 52, 53, 54, 55, 56, 100, 120, 150]),
            1.82, places=2
        )

        # default bins
        self.assertAlmostEqual(entropy(x), 2.91, places=2)


if __name__ == '__main__':
    unittest.main()



import os
import sys
import unittest

import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
from sklearn.model_selection import GridSearchCV

from skgstat import Variogram, OrdinaryKriging
from skgstat.interfaces import VariogramEstimator


def get_sample() -> pd.DataFrame:
    df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'pan_sample.csv'))
    return df


try:
    import pykrige
    print(f'Found PyKrige: {pykrige.__version__}')
    from skgstat.interfaces import pykrige as pykrige_interface
    PYKRIGE_AVAILABLE = True
except ImportError:
    PYKRIGE_AVAILABLE = False  # pragma: no cover

try:
    import gstools
    print(f'Found PyKrige: {gstools.__version__}')
    GSTOOLS_AVAILABLE = True
except ImportError:
    GSTOOLS_AVAILABLE = False  # pragma: no cover


class TestVariogramEstimator(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.c = np.random.normal(50, 8, (300, 2))
        np.random.seed(42)
        self.v = np.random.gamma(10, 1, 300)

    def test_default(self):
        """
        Test default instantition by comparing the fitted
        bins and experimental variogram values to expected arrays.

        """
        # test the built Variogram against bins and experimental
        ve = VariogramEstimator(n_lags=8, normalize=False)

        # fit, to obtain Variogram
        ve_fit = ve.fit(self.c, self.v)
        V = ve_fit.variogram

        assert_array_almost_equal(
            V.bins,
            [6.67, 13.35, 20.02, 26.7, 33.37, 40.04, 46.72, 53.39],
            decimal=2
        )

        assert_array_almost_equal(
            V.experimental,
            [11.01, 10.47, 9.84, 9.17, 8.5, 6.64, 7.41, 4.54],
            decimal=2
        )

    def test_predict_func(self):
        """
        Test predict function. should return the same thing as the
        transform function of the underlying variogram.

        """
        ve = VariogramEstimator(n_lags=15, normalize=False).fit(self.c, self.v)
        v = ve.variogram

        x = np.linspace(0, ve.range_, 100)

        assert_array_almost_equal(ve.predict(x), v.transform(x), decimal=6)

    def test_find_best_model(self):
        """
        Use GridSearchCV to find the best model for the given data
        which should be the spherical model
        """
        parameters = dict(
            model=('spherical', 'gaussian', 'exponential', 'matern')
        )
        gs = GridSearchCV(
            VariogramEstimator(n_lags=15, normalize=False),
            parameters,
            cv=3
        )

        gs = gs.fit(self.c, self.v)

        # Python 3.6 yields 'exponential', 
        # while 3.7, 3.8 yield 'gaussian' - this is so stupid
        self.assertTrue(gs.best_params_['model'] in ['gaussian', 'exponential'])

    def test_find_best_model_future_cv(self):
        """
        cv parameter will change to 5 in sklearn 0.22
        This will change the result, though
        """
        parameters = dict(
            model=('spherical', 'gaussian', 'exponential', 'matern')
        )
        gs = GridSearchCV(
            VariogramEstimator(n_lags=15, normalize=False),
            parameters,
            cv=5
        )

        gs = gs.fit(self.c, self.v)

        self.assertEqual(gs.best_params_['model'], 'exponential')

    def test_cross_validation_option(self):
        # this test does not support python < 3.8
        # the code works, but GridSearchCV gives dififerent results
        if sys.version_info[1] < 8:
            return True

        df = get_sample()
        c = df[['x', 'y']].values
        v = df.z.values

        parameters = dict(
            bin_func=('even', 'scott', 'sqrt')
        )

        # based on model fit
        gs_nocv = GridSearchCV(
            VariogramEstimator(model='exponential', maxlag=100, n_lags=25, cross_n=50),
            parameters,
            cv=3
        )
        gs_nocv = gs_nocv.fit(c, v)
        self.assertEqual(gs_nocv.best_params_['bin_func'], 'sqrt')

        # based on cross-validation
        gs = GridSearchCV(
            VariogramEstimator(
                model='exponential',
                maxlag=100,
                n_lags=25,
                cross_n=50,
                cross_validate=True,
                use_score='mae',
                seed=42
                ),
            parameters,
            cv=3
        )
        gs = gs.fit(c, v)
        # somehow the Py 3.8 and 3.7 differ here, and I don't know why
        self.assertTrue(gs.best_params_['bin_func'] in ('even', 'scott'))


class TestPyKrigeInterface(unittest.TestCase):
    def setUp(self):
        # use real sample data in the interface
        df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'sample.csv'))
        self.c = df[['x', 'y']].values
        self.v = df.z.values

        self.V = Variogram(self.c, self.v, model='matern', normalize=False, use_nugget=True)

        if not PYKRIGE_AVAILABLE:
            print('PyKrige not found, will skip all pykrige interface tests')

    def test_model_interface(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True
        # get the function
        model = pykrige_interface.pykrige_model(self.V)

        # use the transform function.
        xi = np.arange(1, 85)
        yi = self.V.transform(xi)

        assert_array_almost_equal(yi, model([], xi), decimal=6)

    def test_model_interface_from_list(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True

        # get the function
        model = pykrige_interface.pykrige_model(self.V)

        # use the transform function
        xi = list(range(1, 85))
        yi = self.V.transform(np.array(xi))

        assert_array_almost_equal(yi, model([], xi), decimal=6)

    def test_parameters(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True

        p = pykrige_interface.pykrige_params(self.V)
        params = self.V.parameters

        self.assertAlmostEqual(p[0], params[1], places=4)
        self.assertAlmostEqual(p[1], params[0], places=4)
        self.assertAlmostEqual(p[2], params[2], places=4)

    def test_as_kwargs(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True

        args = pykrige_interface.pykrige_as_kwargs(self.V)
        pars = pykrige_interface.pykrige_params(self.V)

        # test
        self.assertEqual(args['variogram_model'], 'custom')
        assert_array_almost_equal(pars, args['variogram_parameters'])

        xi = np.arange(1, 80)
        yi = self.V.transform(xi)
        assert_array_almost_equal(
            yi,
            args['variogram_function']([], xi),
            decimal=6
        )

    def test_as_kwargs_adjust_maxlag(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True

        V = self.V.clone()

        # now maxlag should be changed
        args = pykrige_interface.pykrige_as_kwargs(V, adjust_maxlag=True)

        # should be None
        self.assertIsNone(V.maxlag)

        # transform should change
        xi = np.arange(1, 20)
        yi = V.transform(xi)

        # test changed values
        assert_array_almost_equal(yi, args['variogram_function']([], xi))

    def test_as_kwargs_adjust_nlags(self):
        if not PYKRIGE_AVAILABLE:  # pragma: no cover
            return True

        args = pykrige_interface.pykrige_as_kwargs(self.V, adjust_nlags=True)

        self.assertEqual(args['nlags'], self.V.n_lags)


class TestGstoolsInterface(unittest.TestCase):
    def setUp(self):
        # use real sample data in the interface
        df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'sample.csv'))
        self.c = df[['x', 'y']].values
        self.v = df.z.values

        # build variogram
        self.V = Variogram(self.c, self.v, model='stable', normalize=False, use_nugget=True)

        # model data
        self.xi = np.linspace(0, self.V.bins[-1], 100)
        self.yi = self.V.transform(self.xi)

        if not GSTOOLS_AVAILABLE:
            print('GSTools not found, will skip all gstools interface tests')

    def test_interface(self):
        if not GSTOOLS_AVAILABLE:  # pragma: no cover
            return True

        model = self.V.to_gstools(dim=2)

        assert_array_almost_equal(
            model.variogram(self.xi), self.yi, decimal=2
        )

    def test_infer_dims(self):
        if not GSTOOLS_AVAILABLE:  # pragma: no cover
            return True

        model = self.V.to_gstools()

        assert_array_almost_equal(
            model.variogram(self.xi), self.yi, decimal=2
        )


class TestGstoolsAllModels(unittest.TestCase):
    def setUp(self):
        # use real sample data in the interface
        df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'sample.csv'))
        self.c = df[['x', 'y']].values
        self.v = df.z.values

        if not GSTOOLS_AVAILABLE:
            print('GSTools not found, will skip all gstools interface tests')

    def assert_model(self, model: str):
        if not GSTOOLS_AVAILABLE:  # pragma: no cover
            return True

        # build the model
        V = Variogram(self.c, self.v, model=model, normalize=False)

        # model data
        xi = np.linspace(0, V.bins[-1], 100)
        yi = V.transform(xi)

        model = V.to_gstools()

        assert_array_almost_equal(
            model.variogram(xi), yi, decimal=2
        )

    def test_stable_model(self):
        self.assert_model('stable')

    def test_spherical_model(self):
        self.assert_model('spherical')

    def test_exponential_model(self):
        self.assert_model('exponential')

    def test_cubic_model(self):
        self.assert_model('cubic')

    def test_gaussian_model(self):
        self.assert_model('gaussian')

    def test_matern_model(self):
        self.assert_model('matern')


class TestGstoolsKrige(unittest.TestCase):
    def setUp(self):
        # use real sample data in the interface
        df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'sample.csv'))
        self.c = df[['x', 'y']].values
        self.v = df.z.values

        # build variogram
        self.V = Variogram(self.c[1:], self.v[1:], model='stable', normalize=False, use_nugget=False)

    def test_ordinary(self):
        if not GSTOOLS_AVAILABLE:  # pragma: no cover
            return True

        x = np.array([self.c[0][0]])
        y = np.array([self.c[0][1]])

        # run ordinary kriging with skgstat
        ok = OrdinaryKriging(self.V, min_points=3)
        sk_res = ok.transform(x, y)

        # get the gstools Krige class
        krige = self.V.to_gs_krige()
        gs_res, _ = krige.structured([x, y])

        # test
        assert_array_almost_equal(
            sk_res.flatten(),
            gs_res.flatten(),
            decimal=1
        )


if __name__ == '__main__':
    os.environ['SKG_SUPRESS'] = 'TRUE'  # pragma: no cover
    unittest.main()  # pragma: no cover



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('.'))


def get_version():
    B = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(B, '..', 'VERSION'), 'r') as f:
        return f.read().strip()

# -- Project information -----------------------------------------------------


project = 'SciKit GStat'
copyright = '2021, Mirko Mälicke'
author = 'Mirko Mälicke'

# The short X.Y version
# version = '0.3.2'
# The full version, including alpha/beta/rc tags
release = get_version()


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    'IPython.sphinxext.ipython_console_highlighting',
    'IPython.sphinxext.ipython_directive',
    'nbsphinx',
    'sphinx_gallery.load_style'
#    'numpydoc'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ['.rst', '.ipynb']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'github_url': 'https://github.com/mmaelicke/scikit-gstat',
}

html_context = {
    'github_user': 'mmaelicke',
    'github_repo': 'scikit-gstat',
    'github_version': 'master',
    'doc_path': 'docs'
}

html_short_title = 'SciKit-GStat'
"""
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
        'donate.html'
    ]
}
"""

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'SciKitGStatdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'SciKitGStat.tex', 'SciKit GStat Documentation',
     'Mirko Mälicke', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'scikitgstat', 'SciKit GStat Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'SciKitGStat', 'SciKit GStat Documentation',
     author, 'SciKitGStat', 'One line description of project.',
     'Miscellaneous'),
]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {'https://docs.python.org/': None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Intersphinx mapping -----------------------------------------------------
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.6', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'numpy':  ('https://docs.scipy.org/doc/numpy', None),
    'scipy':  ('https://docs.scipy.org/doc/scipy/reference', None),
    'gstools': ('https://geostat-framework.readthedocs.io/projects/gstools/en/latest/', None),
    'sklearn': ('http://scikit-learn.org/stable', None),
}



import pytest
from skgstat.plotting import backend

import matplotlib.pyplot as plt
import plotly.graph_objects as go 


def test_backend_no_args():
    """
    The default backend should be 'matplotlib'
    """
    assert backend() == 'matplotlib'


@pytest.mark.depends(on=['test_backend_no_args'])
def test_raise_value_error():
    """
    Raise a value error by setting the wrong backend
    """
    with pytest.raises(ValueError):
        backend('not-a-backend')


@pytest.mark.depends(on=['test_raise_value_error'])
def test_change_plotting_backend():
    """
    Set the correct backend and check
    """
    # change to plotly
    backend('plotly')
    assert backend() == 'plotly'

    # change back
    backend('matplotlib')



import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal
import matplotlib.pyplot as plt

from skgstat import SpaceTimeVariogram


class TestSpaceTimeVariogramInitialization(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.c = np.random.gamma(10, 6, (60, 3))
        np.random.seed(42)
        self.v = np.random.normal(15, 4, (60, 8))

    def test_default_init(self):
        V = SpaceTimeVariogram(self.c, self.v)

        # test first 5 and
        assert_array_almost_equal(
            V.experimental[:5],
            np.array([14.527, 16.275, 16.195, 14.464, 12.619]),
            decimal=3
        )

        # and last 5
        assert_array_almost_equal(
            V.experimental[-5:],
            np.array([13.911, 10.76 , 10.623,  9.434, 15.402]),
            decimal=3
        )

    def test_values_setter(self):
        V = SpaceTimeVariogram(self.c, self.v)

        # get the differences
        diff = V.values

        # delete and reset by setter
        V._values = None
        self.assertIsNone(V.values)
        V.values = self.v

        # assert
        assert_array_almost_equal(V.values, diff, decimal=5)

    def test_set_values_raises_AttributeError(self):
        V = SpaceTimeVariogram(self.c, self.v)

        with self.assertRaises(AttributeError) as e:
            V.set_values(['string', 'don\'t', 'work'])
            self.assertEqual(
                str(e),
                'values cannot be converted to a proper '
                '(m,n) shaped array.'
            )

    def test_set_values_raises_shape_error(self):
        V = SpaceTimeVariogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_values(np.random.normal(10, 5, (55, 8)))
            self.assertEqual(
                str(e), 'The values shape do not match coordinates.'
            )

    def test_set_value_raises_timeseries_error(self):
        V = SpaceTimeVariogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_values(np.random.normal(0, 1, (60, 1)))
            self.assertEqual(
                str(e),
                'A SpaceTimeVariogram needs more than one '
                'observation on the time axis.'
            )


class TestSpaceTimeVariogramArgumets(unittest.TestCase):
    def setUp(self):
        np.random.seed(1306)
        self.c = np.random.gamma(20, 10, (60, 3))
        np.random.seed(1306)
        self.v = np.random.power(5, (60, 5))

    def test_xdist_func(self):
        # use Manhattan distance
        V = SpaceTimeVariogram(self.c, self.v, xdist_func='cityblock')

        self.assertEqual(V.xdistance.size, 1770)
        # test arbitrary elements
        assert_array_almost_equal(
            V.xdistance[[10,  15, 492, 1023, 1765]],
            np.array([184.245, 162.17 ,  46.296, 138.417,  91.457]),
            decimal=3
        )

    def test_xdist_func_raises_ValueError(self):
        with self.assertRaises(ValueError) as e:
            V = SpaceTimeVariogram(self.c, self.v)
            V.xdist_func = lambda x: x**2

            self.assertEqual(
                str(e), 'For now only str arguments are supported.'
            )

    def test_tdist_func(self):
        V = SpaceTimeVariogram(self.c, self.v, tdist_func='jaccard')

        # with jaccard, all shoud disagree
        self.assertTrue(all([_ == 1. for _ in V.tdistance]))

    def test_tdist_func_raises_ValueError(self):
        with self.assertRaises(ValueError) as e:
            V = SpaceTimeVariogram(self.c, self.v)
            V.tdist_func = 55.4

            self.assertEqual(
                str(e), 'For now only str arguments are supported.'
            )

    def test_x_lags(self):
        V = SpaceTimeVariogram(self.c, self.v)

        self.assertEqual(V.x_lags, 10)
        V.x_lags = 25
        self.assertEqual(len(V.xbins), 25)

    def test_x_lags_raises_ValueError(self):
        with self.assertRaises(ValueError) as e:
            SpaceTimeVariogram(self.c, self.v, x_lags=15.4)

            self.assertEqual(
                str(e), 'Only integers are supported as lag counts.'
            )

    def test_t_lags(self):
        V = SpaceTimeVariogram(self.c, self.v, t_lags=2)

        self.assertEqual(V.t_lags, 2)
        self.assertEqual(len(V.tbins), 2)
        V.t_lags = 'max'
        # this is still a bug, needs to be fixed one day
        self.assertEqual(V.t_lags, 4)
        self.assertEqual(len(V.tbins), 4)

    def test_t_lags_unkown(self):
        with self.assertRaises(ValueError) as e:
            SpaceTimeVariogram(self.c, self.v, t_lags='min')

            self.assertEqual(
                str(e), "Only 'max' supported as string argument."
            )

    def test_autoset_lag_bins(self):
        V = SpaceTimeVariogram(self.c, self.v, xbins='scott', tbins='fd')

        # test if the bins were set correctly
        self.assertTrue(V.x_lags == 20)
        self.assertTrue(V.t_lags == 2)

        assert_array_almost_equal(
            V.xbins,
            np.array([21.5, 34.8, 48., 61.3, 74.5, 87.8, 101.1, 114.3, 127.6,
                140.8, 154.1, 167.4, 180.6, 193.9, 207.2, 220.4, 233.7, 246.9,
                260.2, 273.5]),
                decimal=1
        )

    def test_change_lag_method(self):
        V = V = SpaceTimeVariogram(self.c, self.v, x_lags=4, t_lags='max')

        self.assertTrue(V.x_lags == V.t_lags == 4)

        V.set_bin_func('sqrt', 'space')

        self.assertTrue(V.x_lags == 43)


class TestSpaceTimeVariogramPlots(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.c = np.random.gamma(14, 8, (50,3))
        np.random.seed(42)
        self.v = np.random.normal(10, 5, (50, 7))

    def test_plot3d_scatter_default(self):
        pass

    def test_plot3d_wrong_axis(self):
        with self.assertRaises(ValueError) as e:
            SpaceTimeVariogram(self.c, self.v)._plot3d(ax=[plt.figure(), 55])

            self.assertEqual(
                str(e),
                'The passed ax object is not an instance '
                'of mpl_toolkis.mplot3d.Axes3D.'
            )


if __name__ == '__main__':
    unittest.main()



import numpy as np


def shannon_entropy(x, bins):
    """Shannon Entropy

    Calculates the Shannon Entropy, which is the most basic
    metric in information theory. It can be used to calculate
    the information content of discrete distributions.
    This can be used to estimate the intrinsic uncertainty of
    a sample,independent of the value range or variance, which
    makes it more comparable.

    Parameters
    ----------
    x : numpy.ndarray
        flat 1D array of the observations
    bins : list, int
        upper edges of the bins used to calculate the histogram
        of x.
    
    Returns
    -------
    h : float
        Shannon Entropy of x, given bins.
    """
    # histogram
    c, _ = np.histogram(x, bins=bins)

    # empirical probabilities
    p = c / np.sum(c) + 1e-15

    # map information function and return product
    return - np.fromiter(map(np.log2, p), dtype=float).dot(p)



import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from skgstat.binning import (
    even_width_lags,
    uniform_count_lags,
    auto_derived_lags,
    kmeans,
    ward,
    stable_entropy_lags
)


class TestEvenWidth(unittest.TestCase):
    @staticmethod
    def test_normal():
        np.random.seed(42)
        bins, _ = even_width_lags(np.random.normal(5, 1, 1000), 4, None)

        assert_array_almost_equal(
            bins,
            np.array([2.21318287, 4.42636575, 6.63954862, 8.85273149])
        )

    @staticmethod
    def test_more_bins():
        np.random.seed(42)

        bins, _ = even_width_lags(np.random.normal(5, 1, 1000), 10, None)

        assert_array_almost_equal(
            bins,
            np.array([0.88527315, 1.7705463, 2.65581945, 3.5410926, 4.42636575,
                      5.3116388, 6.19691204, 7.08218519, 7.96745834, 8.8527314])
        )

    @staticmethod
    def test_maxlag():
        np.random.seed(42)
        bins, _ = even_width_lags(np.random.normal(5, 1, 1000), 4, 4.4)

        assert_array_almost_equal(
            bins,
            np.array([1.1, 2.2, 3.3, 4.4])
        )

    @staticmethod
    def test_too_large_maxlag():
        np.random.seed(42)

        bins, n = even_width_lags(np.random.normal(5, 1, 1000), 4, 400)
        assert_array_almost_equal(
            bins,
            np.array([2.21318287, 4.42636575, 6.63954862, 8.85273149])
        )

    @staticmethod
    def test_median_split():
        np.random.seed(42)
        bins, _ = even_width_lags(np.random.normal(5, 1, 1000), 2, None)
        assert_array_almost_equal(
            bins,
            np.array([4.42636575, 8.85273149])
        )


class TestUniformCount(unittest.TestCase):
    def test_normal(self):
        np.random.seed(42)

        bins, _ = uniform_count_lags(np.random.normal(10, 2, 1000), 4, None)
        assert_array_almost_equal(
            bins,
            np.array([8.7048, 10.0506, 11.2959, 17.7055]),
            decimal=4
        )


class TestDerivedBins(unittest.TestCase):
    def test_auto(self):
        np.random.seed(42)

        bins, n = auto_derived_lags(
            np.random.normal(10, 2, 1000),
            'sturges',
            None
        )

        # sturges should find 11 classes
        self.assertTrue(n == 11)

        assert_array_almost_equal(
            bins,
            np.array([4.8, 6.1, 7.4, 8.7, 10., 11.3, 12.5, 13.8, 15.1, 16.4, 17.7]),
            decimal=1
        )

    def test_skewed(self):
        np.random.seed(1312)

        bins, n = auto_derived_lags(np.random.gamma(10, 20, 500), 'doane', 100)

        # doane should condense to 6 here
        self.assertTrue(n == 6)

        assert_array_almost_equal(
            bins,
            np.array([75.6, 80.4, 85.3, 90.2, 95.1, 100.]),
            decimal=1
        )


class TestClusteringBins(unittest.TestCase):
    def test_kmeans(self):
        np.random.seed(1312)
        bins, _ = kmeans(np.random.gamma(10, 40, 500), 6, None)

        assert_array_almost_equal(
            np.array([118.5, 283.2, 374.7, 467.9, 574.5, 762.5]),
            bins,
            decimal=1
        )

    def test_ward(self):
        np.random.seed(1312)
        bins, _ = ward(np.random.gamma(10, 40, 500), 6, None)

        assert_array_almost_equal(
            np.array([122.7, 283.7, 352.3, 422.7, 520.9, 660.4]),
            bins,
            decimal=1
        )

    def test_ward_median(self):
        np.random.seed(1312)
        bins, _ = ward(np.random.gamma(10, 40, 500), 6, None, binning_agg_func='median')

        assert_array_almost_equal(
            np.array([126.2, 287.8, 354.6, 421.7, 517., 643.1]),
            bins,
            decimal=1
        )

    def test_stable_entropy(self):
        np.random.seed(1312)
        d = np.random.gamma(1500, 40, 100)

        # run
        bins, _ = stable_entropy_lags(d, 6, None)

        assert_array_almost_equal(
            np.array([10964.6, 21929.3, 32893.9, 43858.6, 54823.2, 69077.2]),
            bins,
            decimal=1
        )


if __name__ == '__main__':
    unittest.main()



"""
from skgstat.tests.estimator import TestEstimator
from skgstat.tests.models import TestModels, TestVariogramDecorator
from skgstat.tests.binning import TestEvenWidth, TestUniformCount
from skgstat.tests.Variogram import (
    TestVariogramInstatiation,
    TestVariogramArguments,
    TestVariogramFittingProcedure,
    TestVariogramQaulityMeasures,
    TestVariogramMethods,
    TestVariogramPlots,
)
from skgstat.tests.DirectionalVariogram import (
    TestDirectionalVariogramInstantiation,
    TestDirectionalVariogramMethods,
)
from skgstat.tests.SpaceTimeVariogram import (
    TestSpaceTimeVariogramInitialization,
    TestSpaceTimeVariogramArgumets,
    TestSpaceTimeVariogramPlots,
)
from skgstat.tests.kriging import (
    TestKrigingInstantiation,
    TestPerformance,
)
from skgstat.tests.interfaces import (
    TestVariogramEstimator,
    TestPyKrigeInterface,
    TestGstoolsInterface
)
from skgstat.tests.stmodels import (
    TestSumModel,
    TestProductModel,
    TestProductSumModel
)

import os
os.environ['SKG_SUPRESS'] = 'TRUE'
"""


"""GSTools Interface."""
import numpy as np


def stable_rescale(describe):
    """Get GSTools rescale parameter from sk-gstat stable model description."""
    return np.power(3, 1 / describe["shape"])


MODEL_MAP = dict(
    spherical=dict(gs_cls="Spherical"),
    exponential=dict(gs_cls="Exponential", rescale=3.0),
    gaussian=dict(gs_cls="Gaussian", rescale=2.0),
    cubic=dict(gs_cls="Cubic"),
    stable=dict(
        gs_cls="Stable", arg_map={"alpha": "shape"}, rescale=stable_rescale
    ),
    matern=dict(
        gs_cls="Matern", arg_map={"nu": "smoothness"}, rescale=4.0
    ),
)


def skgstat_to_gstools(variogram, **kwargs):
    """
    Instantiate a corresponding GSTools :any:`CovModel <gstools.CovModel>`.

    By default, this will be an isotropic model.

    Parameters
    ----------
    variogram : :any:`Variogram`
        Scikit-Gstat Variogram instance.
    **kwargs
        Keyword arguments forwarded to the instantiated GSTools CovModel.
        The default parameters 'dim', 'var', 'len_scale', 'nugget',
        'rescale' and optional shape parameters will be extracted
        from the given Variogram but they can be overwritten here.

    Raises
    ------
    ImportError
        When GSTools is not installed.
    ValueError
        When GSTools version is not v1.3 or greater.
    ValueError
        When given Variogram model is not supported ('harmonize').

    Returns
    -------
    model : :any:`CovModel <gstools.CovModel>`
        Corresponding GSTools covmodel.

    Note
    ----
    In case you intend to use the
    :func:`coordinates <skgstat.Variogram.coordinates>`
    in a GSTools workflow, you need to transpose the coordinate
    array like:

    >> cond_pos Variogram.coordinates.T

    """
    # try to import gstools and notify user if not installed
    try:
        import gstools as gs
    except ImportError as e:  # pragma: no cover
        raise ImportError("to_gstools: GSTools not installed.") from e

    # at least gstools>=1.3.0 is needed
    if list(map(int, gs.__version__.split(".")[:2])) < [1, 3]:  # pragma: no cover
        raise ValueError("to_gstools: GSTools v1.3 or greater requiered.")

    # gstolls needs the spatial dimension
    kwargs.setdefault("dim", variogram.dim)

    # extract all needed settings
    describe = variogram.describe()

    # get the theoretical model name
#    name = describe["name"]
    name = describe['model']

    if name not in MODEL_MAP:
        raise ValueError("skgstat_to_gstools: model not supported: " + name)
    gs_describe = MODEL_MAP[name]

    # set variogram parameters
    gs_describe.setdefault("rescale", 1.0)
    gs_describe.setdefault("arg_map", dict())
    gs_kwargs = dict(
        var=float(describe["sill"] - describe["nugget"]),
        len_scale=float(describe["effective_range"]),
        nugget=float(describe["nugget"]),
    )

    # some skgstat models need different rescale
    rescale = gs_describe["rescale"]
    gs_kwargs["rescale"] = rescale(describe) if callable(rescale) else rescale
    arg_map = gs_describe["arg_map"]
    for arg in arg_map:
        gs_kwargs[arg] = float(describe[arg_map[arg]])

    # update the parameters
    gs_kwargs.update(kwargs)

    # get the model and return the CovModel
    gs_model = getattr(gs, gs_describe["gs_cls"])
    return gs_model(**gs_kwargs)


def skgstat_to_krige(variogram, **kwargs):
    """
    Instatiate a GSTools Krige class.

    This can only export isotropic models.
    Note: the `fit_variogram` is always set to `False`

    Parameters
    ----------
    variogram : skgstat.Variogram
        Scikit-GStat Variogram instamce
    **kwargs
        Keyword arguments forwarded to GSTools Krige.
        Refer to :any:`Krige <gstools.krige.Krige>` to
        learn about all possible options.
        Note that the `fit_variogram` parameter will
        always be False.

    Raises
    ------
    ImportError
        When GSTools is not installed.
    ValueError
        When GSTools version is not v1.3 or greater.
    ValueError
        When given Variogram model is not supported ('harmonize').

    Returns
    -------
    :any:`Krige <gstools.krige.Krige>`
        Instantiated GSTools Krige class.

    See Also
    --------
    gstools.Krige

    """
    # try to import gstools and notify user if not installed
    try:
        import gstools as gs
    except ImportError as e:  # pragma: no cover
        raise ImportError("to_gstools: GSTools not installed.") from e

    # at least gstools>=1.3.0 is needed
    if list(map(int, gs.__version__.split(".")[:2])) < [1, 3]:  # pragma: no cover
        raise ValueError("to_gstools: GSTools v1.3 or greater requiered.")

    # convert variogram to a CovModel
    model = skgstat_to_gstools(variogram=variogram)

    # extract cond_pos and cond_vals
    cond_pos = variogram.coordinates.T
    cond_vals = variogram.values

    # disable the re-fitting of the variogram in gstools
    kwargs['fit_variogram'] = False

    # instantiate the Krige class
    krige = gs.krige.Krige(model, cond_pos, cond_vals, **kwargs)

    # return the class
    return krige



import math
from functools import wraps

import numpy as np
from scipy import special
from numba import jit


def variogram(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if hasattr(args[0], '__iter__'):
            new_args = args[1:]
            mapping = map(lambda h: func(h, *new_args, **kwargs), args[0])
            return np.fromiter(mapping, dtype=float)
        else:
            return func(*args, **kwargs)
    return wrapper


@variogram
@jit
def spherical(h, r, c0, b=0):
    r"""Spherical Variogram function

    Implementation of the spherical variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! However,
        for the spherical variogram the range and effective range are the same.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----

    The implementation follows [6]_:

    .. math::
        \gamma = b + C_0 * \left({1.5*\frac{h}{a} - 0.5*\frac{h}{a}^3}\right)

    if :math:`h < r`, and

    .. math::
        \gamma = b + C_0

    else. r is the effective range, which is in case of the spherical
    variogram just a.

    References
    ----------

    .. [6] Burgess, T. M., & Webster, R. (1980). Optimal interpolation
       and isarithmic mapping of soil properties. I.The semi-variogram and
       punctual kriging. Journal of Soil and Science, 31(2), 315–331,
       http://doi.org/10.1111/j.1365-2389.1980.tb02084.x

    """
    # prepare parameters
    a = r / 1.

    if h <= r:
        return b + c0 * ((1.5 * (h / a)) - (0.5 * ((h / a) ** 3.0)))
    else:
        return b + c0


@variogram
@jit
def exponential(h, r, c0, b=0):
    r"""Exponential Variogram function

    Implementation of the exponential variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! For the
        exponential variogram function the range parameter a is defined to be
        :math:`a=\frac{r}{3}`. The effective range is the lag where 95% of the
        sill are exceeded. This is needed as the sill is only approached
        asymptotically by an exponential function.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----
    The implementation following [7]_, [9]_ and [8]_ is as:

    .. math::
        \gamma = b + C_0 * \left({1 - e^{-\frac{h}{a}}}\right)

    a is the range parameter, that can be calculated from the
    effective range r as: :math:`a = \frac{r}{3}`.

    References
    ----------

    .. [7] Cressie, N. (1993): Statistics for spatial data.
       Wiley Interscience.

    .. [8] Chiles, J.P., Delfiner, P. (1999). Geostatistics. Modeling Spatial
       Uncertainty. Wiley Interscience.

    .. [9] Journel, A G, and Huijbregts, C J. Mining geostatistics. 
        United Kingdom: N. p., 1976.

    """
    # prepare parameters
    a = r / 3.

    return b + c0 * (1. - math.exp(-(h / a)))


@variogram
@jit
def gaussian(h, r, c0, b=0):
    r""" Gaussian Variogram function

    Implementation of the Gaussian variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! For the
        exponential variogram function the range parameter a is defined to be
        :math:`a=\frac{r}{3}`. The effetive range is the lag where 95% of the
        sill are exceeded. This is needed as the sill is only approached
        asymptotically by an exponential function.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----

    This implementation follows [10]_ and [11]_:

    .. math::

        \gamma = b + c_0 * \left({1 - e^{-\frac{h^2}{a^2}}}\right)


    a is the range parameter, that can be calculated from the
    effective range r as:

    .. math::
        a = \frac{r}{2}


    References
    ----------

    .. [10] Chiles, J.P., Delfiner, P. (1999). Geostatistics. Modeling Spatial
       Uncertainty. Wiley Interscience.
    .. [11] Journel, A G, and Huijbregts, C J. Mining geostatistics. 
        United Kingdom: N. p., 1976.

    """
    # prepare parameters
    a = r / 2.

    return b + c0 * (1. - math.exp(- (h ** 2 / a ** 2)))


@variogram
@jit
def cubic(h, r, c0, b=0):
    r"""Cubic Variogram function

    Implementation of the Cubic variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! However,
        for the cubic variogram the range and effective range are the same.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----

    This implementation is taken from [12]_:

    .. math::

        \gamma = b + C_0 *  \left[{7 * \left(\frac{h^2}{a^2}\right) -
        \frac{35}{4} * \left(\frac{h^3}{a^3}\right) +
        \frac{7}{2} * \left(\frac{h^5}{a^5}\right) -
        \frac{3}{4} * \left(\frac{h^7}{a^7}\right)}\right]

    a is the range parameter. For the cubic function, the effective range and
    range parameter are the same.

    References
    ----------

    .. [12] Montero, J.-M., Mateu, J., & others. (2015). Spatial and spatio-temporal 
        geostatistical modeling and kriging (Vol. 998). John Wiley & Sons.

    """
    # prepare parameters
    a = r / 1.

    if h < r:
        return b + c0 * ((7 * (h ** 2 / a ** 2)) -
                         ((35 / 4) * (h ** 3 / a ** 3)) +
                         ((7 / 2) * (h ** 5 / a ** 5)) -
                         ((3 / 4) * (h ** 7 / a ** 7)))
    else:
        return b + c0


@variogram
@jit
def stable(h, r, c0, s, b=0):
    r"""Stable Variogram function

    Implementation of the stable variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    .. versionchanged:: 0.4.4
        Now returns the nugget at lag 0

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! For the
        stable variogram function the range parameter a is defined to be
        :math:`a = \frac{r}{3^{\frac{1}{s}}}`. The effective range is the lag
        where 95% of the sill are exceeded. This is needed as the sill is
        only approached asymptotically by the e-function part of the stable
        model.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    s : float
        Shape parameter. For s <= 2 the model will be shaped more like a
        exponential or spherical model, for s > 2 it will be shaped most like
        a Gaussian function.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----
    The implementation is taken from [12]_:

    .. math::
        \gamma = b + C_0 * \left({1. - e^{- {\frac{h}{a}}^s}}\right)

    a is the range parameter and is calculated from the effective range r as:

    .. math::
        a = \frac{r}{3^{\frac{1}{s}}}

    References
    ----------

    .. [12] Montero, J.-M., Mateu, J., & others. (2015). Spatial and spatio-temporal
        geostatistical modeling and kriging (Vol. 998). John Wiley & Sons.

    """
    # if s gts too small, we run into a zeroDivision error at lag 0
    if h == 0:
        return b

    # prepare parameters
    a = r / np.power(3, 1 / s)

#    if s > 2:
#        s = 2
    return b + c0 * (1. - math.exp(- math.pow(h / a, s)))


@variogram
@jit(forceobj=True)
def matern(h, r, c0, s, b=0):
    r"""Matérn Variogram function

    Implementation of the Matérn variogram function. Calculates the
    dependent variable for a given lag (h). The nugget (b) defaults to be 0.

    .. versionchanged:: 0.4.4
        now returns the nugget instead of NaN for lag 0.

    Parameters
    ----------
    h : float
        Specifies the lag of separating distances that the dependent variable
        shall be calculated for. It has to be a positive real number.
    r : float
        The effective range. Note this is not the range parameter! For the
        Matérn variogram function the range parameter a is defined to be
        :math:`a = \frac{r}{2}` and :math:`a = \frac{r}{3}` if s is smaller
        than 0.5 or larger than 10. The effective range is the lag
        where 95% of the sill are exceeded. This is needed as the sill is
        only approached asymptotically by Matérn model.
    c0 : float
        The sill of the variogram, where it will flatten out. The function
        will not return a value higher than C0 + b.
    s : float
        Smoothness parameter. The smoothness parameter can shape a smooth or
        rough variogram function. A value of 0.5 will yield the exponential
        function, while a smoothness of +inf is exactly the Gaussian model.
        Typically a value of 10 is close enough to Gaussian shape to simulate
        its behaviour. Low values are considered to be 'smooth', while larger
        values are considered to describe a 'rough' random field.
    b : float
        The nugget of the variogram. This is the value of independent
        variable at the distance of zero. This is usually attributed to
        non-spatial variance.

    Returns
    -------
    gamma : numpy.float64
        Unlike in most variogram function formulas, which define the function
        for :math:`2*\gamma`, this function will return :math:`\gamma` only.

    Notes
    -----
    The implementation is taken from [13]_:

    .. math::
        \gamma (h) = b + C_0 \left( 1 - \frac{1}{2^{\upsilon - 1}
        \Gamma(\upsilon)}\left(\frac{h}{a}\right)^\upsilon K_\upsilon
        \left(\frac{h}{a}\right)\right)

    a is the range parameter and is calculated from the effective range r as:

    .. math::
        a = \frac{r}{2}

    References
    ----------
    .. [13] Zimmermann, B., Zehe, E., Hartmann, N. K., & Elsenbeer, H. (2008).
        Analyzing spatial data: An assessment of assumptions, new methods, and
        uncertainty using soil hydraulic data. Water Resources Research,
        44(10), 1–18. https://doi.org/10.1029/2007WR006604

    """
    if h == 0:
        return b
    # prepare parameters
    a = r / 2.

    # calculate
    return b + c0 * (1. - (2 / special.gamma(s)) *
                     np.power((h * np.sqrt(s)) / a, s) *
                     special.kv(s, 2 * ((h * np.sqrt(s)) / a))
                     )



"""

"""
import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from skgstat.models import spherical, exponential
from skgstat.models import gaussian, cubic, stable, matern
from skgstat.models import variogram


class TestModels(unittest.TestCase):
    def setUp(self):
        self.h = np.array([5, 10, 30, 50, 100])

    def test_spherical_default(self):
        # extract the actual function
        f = spherical.py_func

        result = [13.75, 20.0, 20.0, 20.0, 20.0]

        model = list(map(f, self.h, [10]*5, [20]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_spherical_nugget(self):
        # extract the actual function
        f = spherical.py_func

        result = [15.44, 27.56, 33.0, 34.0, 35.0]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [15] * 5, [30] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_exponential_default(self):
        # extract the actual function
        f = exponential.py_func

        result = [5.18, 9.02, 16.69, 19., 19.95]
        model = list(map(f, self.h, [50]*5, [20]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_exponential_nugget(self):
        # extract the actual function
        f = exponential.py_func

        result = [7.64, 13.8, 26.31, 31.54, 34.8]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [60] * 5, [30] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_gaussian_default(self):
        # extract the actual function
        f = gaussian.py_func

        result = [0.96,  3.58, 16.62, 19.86, 20.]
        model = list(map(f, self.h, [45]*5, [20]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_gaussian_nugget(self):
        # extract the actual function
        f = gaussian.py_func

        result = [1.82,  5.15, 21.96, 32.13, 35.]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [60] * 5, [30] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def _test_cubic_default(self):
        # extract the actual function
        f = cubic.py_func

        result = [6.13,  21.11,  88.12, 100., 100.]
        model = list(map(f, self.h, [50]*5, [100]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_cubic_nugget(self):
        # extract the actual function
        f = cubic.py_func

        result = [11.81, 34.74, 73., 74., 75.]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [30] * 5, [70] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_stable_default(self):
        # extract the actual function
        f = stable.py_func

        result = [9.05, 23.53, 75.2, 95.02, 99.98]
        model = list(map(f, self.h, [50]*5, [100]*5, [1.5]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_stable_nugget(self):
        # extract the actual function
        f = stable.py_func

        result = [8.77, 10.8, 12.75, 13.91, 14.99]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [20] * 5, [10] * 5, [0.5] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_matern_default(self):
        # extract the actual function
        f = matern.py_func

        result = [24.64, 43.2, 81.68, 94.09, 99.65]
        model = list(map(f, self.h, [50]*5, [100]*5, [0.50001]*5))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)

    def test_matern_nugget(self):
        # extract the actual function
        f = matern.py_func

        result = [3.44, 8.52, 12.99, 14., 15.]

        # calculate
        nuggets = [1, 2, 3, 4, 5]
        model = list(map(f, self.h, [20] * 5, [9.99999] * 5, [8] * 5, nuggets))

        for r, m in zip(result, model):
            self.assertAlmostEqual(r, m, places=2)
    
    def test_matern_r_switch(self):
        # run the default with an extreme s value

        # extract the actual function
        f = matern.py_func

        result = [24.64, 43.20, 81.68, 94.09, 99.65]

        model = list(map(f, self.h, [50]*5, [100]*5, [0.5]*5))

        assert_array_almost_equal(result, model, decimal=2)


class TestVariogramDecorator(unittest.TestCase):
    def test_scalar(self):
        @variogram
        def scalar_function(a, b):
            return a, b

        a, b = 1, 4
        self.assertEqual(scalar_function(1, 4), (a, b))

    def test_list(self):
        @variogram
        def adder(l, a):
            return l + a

        res = [5, 8, 12]

        for r, c in zip(res, adder([1, 4, 8], 4)):
            self.assertEqual(r, c)


if __name__=='__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt

try:
    import plotly.graph_objects as go 
except ImportError:
    pass


def __calculate_plot_data(variogram):
   # get all distances
    _dist = variogram.distance

    # get all differences
    if variogram._diff is None:
        variogram._calc_diff()
    _diff = variogram._diff

    return _diff, _dist


def matplotlib_dd_plot(variogram, ax=None, plot_bins=True, show=True):
    # get the plotting data 
    _diff, _dist = __calculate_plot_data(variogram)
   
    # create the plot
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    else:
        fig = ax.get_figure()

    # plot the bins
    if plot_bins:
        _bins = variogram.bins
        ax.vlines(_bins, 0, np.max(_diff), linestyle='--', lw=1, color='r')

    # plot
    ax.scatter(_dist, _diff, 8, color='b', marker='o', alpha=0.5)

    # set limits
    ax.set_ylim((0, np.max(_diff)))
    ax.set_xlim((0, np.max(_dist)))
    ax.set_xlabel('separating distance')
    ax.set_ylabel('pairwise difference')
    ax.set_title('Pairwise distance ~ difference')

    # show the plot
    if show:  # pragma: no cover
        fig.show()

    return fig


def plotly_dd_plot(variogram, fig=None, plot_bins=True, show=True):
    # get the plotting data
    _diff, _dist = __calculate_plot_data(variogram)

    # create a new Figure if needed
    if fig is None:
        fig = go.Figure()

    # plot
    fig.add_trace(
        go.Scattergl(
            x=_dist, y=_diff, 
            mode='markers', marker=dict(color='blue', opacity=0.5)
        )
    )

    # plot the bins
    if plot_bins:
        for _bin in variogram.bins:
            fig.add_vline(x=_bin, line_dash='dash', line_color='red')

    # titles
    fig.update_layout(
        title='Pairwise distance ~ difference',
        xaxis_title='separating distance',
        yaxis_title='pairwise difference'
    )

    if show:
        fig.show()

    return fig



from typing import Union, Tuple, List
import os
import imageio
from glob import glob
import numpy as np
import pandas as pd


PATH = os.path.abspath(os.path.dirname(__file__))


def field_names() -> List[str]:
    """
    Get all available fields
    """
    fnames = glob(os.path.join(PATH, 'rf', '*.png'))
    basenames = [os.path.basename(name) for name in fnames]
    return [os.path.splitext(name)[0] for name in basenames]


def field(fname: str, band: Union[int, str] = 0) -> np.ndarray:
    """
    Return one of the fields stored in the ``rf`` folder
    and return as numpy array.

    Parameters
    ----------
    fname : str
        The filename (not path) of the field. The file
        extenstion can be omitted.
    band : int, str
        The band to use, can either be an integer or the
        literal ``'mean'``, which will average all bands

    Returns
    -------
    field : numpy.ndarray
        The ndarray representing the requested band

    """
    # append a png file extension if needed
    if not fname.endswith('.png'):
        fname += '.png'

    # read image
    img = imageio.imread(os.path.join(PATH, 'rf', fname))

    # switch band
    if isinstance(band, int):
        if len(img.shape) > 2:
            return img[:, :, band]
        elif len(img.shape) == 2:
            return img
    elif band.lower() == 'mean':
        return np.mean(img, axis=2)

    raise AttributeError('band parameter is invalid')


def get_sample(
        fname: str,
        N: int = 100,
        seed: int = None,
        band: Union[int, str] = 0
) -> Tuple[np.ndarray]:
    """
    Sample one of the fields. The filename and band are passed down to
    :func:`field`. To return reproducible results the random Generator
    can be seeded.

    Parameters
    ----------
    fname : str
        The filename (not path) of the field. The file
        extenstion can be omitted.
    N : int
        Sample size
    seed : int
        seed to use for the random generator.
    band : int, str
        The band to use, can either be an integer or the
        literal ``'mean'``, which will average all bands

    Returns
    -------
    coordinates : numpy.ndarray
        Coordinate array of shape ``(N, 2)``.
    values : numpy.ndarray
        1D array of the values at the coordinates

    """
    # first get the image
    img = field(fname, band)

    # randomly sample points
    rng = np.random.default_rng(seed)

    # sample at random flattened indices without replace
    idx = rng.choice(np.multiply(*img.shape), replace=False, size=N)

    # build a meshgrid over the image
    _x, _y = np.meshgrid(*[range(dim) for dim in img.shape])
    x = _x.flatten()
    y = _y.flatten()

    # get the coordinates and values
    coordinates = np.asarray([[x[i], y[i]] for i in idx])
    values = np.asarray([img[c[0], c[1]] for c in coordinates])

    return coordinates, values


def read_sample_file(fname) -> pd.DataFrame:
    """
    Return a sample from a sample-file as a
    pandas DataFrame

    Returns
    -------
    df : pandas.DataFrame
        The file content

    """
    # build the path
    path = os.path.join(PATH, 'samples', fname)
    return pd.read_csv(path)



try:
    import pykrige
    PYKRIGE_AVAILABLE = True
except ImportError:
    PYKRIGE_AVAILABLE = False

import numpy as np


def __check_pykrige_available():  # pragma: no cover
    if not PYKRIGE_AVAILABLE:
        print('The pykrige interface needs pykrige installed.')
        print("Run 'pip install pykrige' to install it.")
        return False
    return True


def pykrige_model(variogram):
    """
    """
    # pykrige is available?
    if not __check_pykrige_available():
        return

    # get the fitted model
    model = variogram.fitted_model

    # define the model function
    def skgstat_model(parameters, lags):
        """Variogram model

        This function is a interface from scikit-gstat to pykrige.
        If you want to use a fitted skgstat.Variogram instance as a
        custom variogram model in pykrige, this is the already fitted
        function that can be passed as the `variogram_function` argument.
        Additionally, you need to set the `variogram_model` to `'custom'`.

        The skgstat.interfaces.pykrige also has a pykrige_as_kwargs
        function. That will return all necessary keyword arguments for the
        pykrige class as a dictionary. You can just pass it using the double
        star operator.

        """
        if not isinstance(lags, np.ndarray):
            lags = np.asarray(lags, dtype=float)

        # get the semi-variances
        semivar = np.fromiter(map(model, lags.flatten()), dtype=float)

        # return
        return semivar.reshape(lags.shape)

    # return model
    return skgstat_model


def pykrige_params(variogram):
    """
    """
    # pykrige is available?
    if not __check_pykrige_available():
        return

    # get the parameters into the correct order. 
    pars = variogram.parameters

    return [pars[1], pars[0], pars[2]]


def pykrige_as_kwargs(variogram, adjust_maxlag=False, adjust_nlags=False):
    """
    """
    # pykrige is available?
    if not __check_pykrige_available():
        return

    # as far as I get it, there is no maximum lag in pykrige. 
    if adjust_maxlag:
        variogram.maxlag = None
    else:
        print('[WARNING]: If the maximum lag is not None, the variogram plots will differ.')

    # to work properly, variogram has to use a nugget
    variogram.use_nugget = True
    variogram.fit()

    # get the model
    model_func = pykrige_model(variogram)

    # get the parameters
    pars = pykrige_params(variogram)

    args = dict(
        variogram_model='custom',
        variogram_parameters=pars,
        variogram_function=model_func
    )

    if adjust_nlags:
        args['nlags'] = variogram.n_lags

    # return
    return args



"""
This file collects different functions for estimating the semivariance of a group of values. These functions
can be used to fit a experimental variogram to a list of points. Each of the given functions just calculate
the estimate for one bin. If you want a Variogram, use the variogram functions or Class from the Variogram and vario
submodule, or order the bins yourself
"""
import numpy as np
from scipy.special import binom
from numba import njit, jit

from skgstat.util import shannon_entropy


@njit
def matheron(x):
    r"""Matheron Semi-Variance

    Calculates the Matheron Semi-Variance from an array of pairwise differences.
    Returns the semi-variance for the whole array. In case a semi-variance is
    needed for multiple groups, this function has to be mapped on each group.
    That is the typical use case in geostatistics.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.

    Returns
    -------
    numpy.float64

    Notes
    -----

    This implementation follows the original publication [1]_ and the
    notes on their application [2]_. Following the 1962 publication [1]_,
    the semi-variance is calculated as:

    .. math::
        \gamma (h) = \frac{1}{2N(h)} * \sum_{i=1}^{N(h)}(x)^2

    with:

    .. math::
        x = Z(x_i) - Z(x_{i+h})

    where x is exactly the input array x.

    References
    ----------

    .. [1] Matheron, G. (1962): Traité de Géostatistique Appliqué, Tonne 1.
       Memoires de Bureau de Recherches Géologiques et Miniéres, Paris.

    .. [2] Matheron, G. (1965): Les variables regionalisées et leur estimation.
       Editions Masson et Cie, 212 S., Paris.

    """
    # prevent ZeroDivisionError
    if x.size == 0:
        return np.nan

    return (1. / (2 * x.size)) * np.sum(np.power(x, 2))


@njit
def cressie(x):
    r""" Cressie-Hawkins Semi-Variance

    Calculates the Cressie-Hawkins Semi-Variance from an array of pairwise
    differences. Returns the semi-variance for the whole array. In case a
    semi-variance is needed for multiple groups, this function has to be
    mapped on each group. That is the typical use case in geostatistics.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.

    Returns
    -------
    numpy.float64

    Notes
    -----

    This implementation is done after the publication by Cressie and Hawkins
    from 1980 [3]_:

    .. math::
        2\gamma (h) = \frac{(\frac{1}{N(h)} \sum_{i=1}^{N(h)} |x|^{0.5})^4}
        {0.457 + \frac{0.494}{N(h)} + \frac{0.045}{N^2(h)}}

    with:

    .. math::
        x = Z(x_i) - Z(x_{i+h})

    where x is exactly the input array x.

    References
    ----------

    .. [3] Cressie, N., and D. Hawkins (1980): Robust estimation of the
       variogram. Math. Geol., 12, 115-125.


    """
    # get the length
    n = x.size

    # prevent ZeroDivisionError
    if n == 0:
        return np.nan

    # Nominator
    nominator = np.power((1 / n) * np.sum(np.power(x, 0.5)), 4)

    # Denominator
    denominator = 0.457 + (0.494 / n) + (0.045 / n**2)

    return nominator / (2 * denominator)


def dowd(x):
    r"""Dowd semi-variance

    Calculates the Dowd semi-variance from an array of pairwise
    differences. Returns the semi-variance for the whole array. In case a
    semi-variance is needed for multiple groups, this function has to be
    mapped on each group. That is the typical use case in geostatistics.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.

    Returns
    -------
    numpy.float64

    Notes
    -----
    The Dowd estimator is based on the median of all pairwise differences in
    each lag class and is therefore robust to exteme values at the cost of
    variability.
    This implementation follows Dowd's publication [4]_:

    .. math::

        2\gamma (h) = 2.198 * {median(x)}^2

    with:

    .. math::
        x = Z(x_i) - Z(x_{i+h})

    where x is exactly the input array x.

    References
    ----------

    .. [4] Dowd, P. A., (1984): The variogram and kriging: Robust and resistant
       estimators, in Geostatistics for Natural Resources Characterization.
       Edited by G. Verly et al., pp. 91 - 106, D. Reidel, Dordrecht.

    """
    # convert
    return 2.198 * np.nanmedian(x)**2


@jit(forceobj=True)
def genton(x):
    r""" Genton robust semi-variance estimator

    Return the Genton semi-variance of the given sample x. Genton is a highly
    robust varigram estimator, that is designed to be location free and
    robust on extreme values in x.
    Genton is based on calculating kth order statistics and will for large
    data sets be close or equal to the 25% quartile of all ordered point pairs
    in X.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.

    Returns
    -------
    numpy.float64

    Notes
    -----
    The Genton estimator is described in great detail in the original
    publication [5]_ and is defined as:

    .. math:: Q_{N_h} = 2.2191\{|V_i(h) - V_j(h)|; i < j\}_{(k)}

    and

     .. math:: k = \binom{[N_h / 2] + 1}{2}

     and

     .. math:: q = \binom{N_h}{2}

     where k is the kth quantile of all q point pairs. For large N (k/q) will be
     close to 0.25. For N >= 500, (k/q) is close to 0.25 by two decimals and
     will therefore be set to 0.5 and the two binomial coefficients k,
     q are not calculated.

    References
    ----------

    ..  [5] Genton, M. G., (1998): Highly robust variogram estimation,
        Math. Geol., 30, 213 - 221.

    """
    # get length
    n = x.size

    if n < 2:
        return np.nan

    # pre-populate y => we need (n*n -n) / 2
    y = np.zeros(int((n*n - n) / 2))

    # calculate
    z = 0
    for i in range(n):
        for j in range(n):
            if i < j:
                y[z] = np.abs(x[i] - x[j])
                z += 1

    # if N > 500, (k/q) will be ~ 1/4 anyway
    if n >= 500:
        k, q, = 1, 4
    else:
        # get k  k is binom(N(x)/2+1, 2)
        k = binom(n / 2 + 1, 2)

        # get q. Genton needs the kth quantile of q
        q = binom(n, 2)

    # return the kth percentile
    return 0.5 * np.power(2.219 * np.percentile(y, (k / q)), 2)


def minmax(x):
    """Minimum - Maximum Estimator

    Returns a custom value. This estimator is the difference of maximum and
    minimum pairwise differences, normalized by the mean. MinMax will be very
    sensitive to extreme values.

    Do only use this estimator, in case you know what you are doing. It is
    experimental and might change its behaviour in a future version.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.

    Returns
    -------
    numpy.float64

    """
    return (np.nanmax(x) - np.nanmin(x)) / np.nanmean(x)


def percentile(x, p=50):
    """Percentile estimator

    Returns a given percentile as semi-variance. Do only use this estimator,
    in case you know what you are doing.

    Do only use this estimator, in case you know what you are doing. It is
    experimental and might change its behaviour in a future version.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.
    p : int
        Desired percentile. Should be given as whole numbers 0 < p < 100.

    Returns
    -------
    np.float64

    """
    return np.percentile(x, q=p)


def entropy(x, bins=None):
    """Shannon Entropy estimator

    Calculates the Shannon Entropy H as a variogram estimator. It is highly
    recommended to calculate the bins and explicitly set them as a list.
    In case this function is called for more than one lag class in a
    variogram, setting bins to None would result in different bin edges in
    each lag class. This would be very difficult to interpret.

    Parameters
    ----------
    x : numpy.ndarray
        Array of pairwise differences. These values should be the distances
        between pairwise observations in value space. If xi and x[i+h] fall
        into the h separating distance class, x should contain abs(xi - x[i+h])
        as an element.
    bins : int, list, str
        list of the bin edges used to calculate the empirical distribution of x.
        If bins is a list, these values are used directly. In case bins is a
        integer, as many even width bins will be calculated between the
        minimum and maximum value of x. In case bins is a string, it will be
        passed as bins argument to numpy.histograms function.

    Returns
    -------
    entropy : numpy.float64
        Shannon entropy of the given pairwise differences.

    Notes
    -----

    """
    # this is actually a bad idea
    if bins is None:
        bins = 15

    return shannon_entropy(x, bins)



import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal

from skgstat import models, stmodels


class TestSumModel(unittest.TestCase):
    def setUp(self):
        # spatial range = 10;  spatial sill = 5
        self.Vx = lambda h: models.spherical(h, 10, 5)
        # temporal range = 5;  temporal sill = 7
        self.Vt = lambda t: models.cubic(t, 5, 7)

        self.lags = np.array([
            [1.2, 1.0],
            [5.0, 2.5],
            [10., 5.0],
            [12., 7.0],
            [4.2, 7.0],
            [12.0, 3.4]
        ])

    def test_default(self):
        assert_array_almost_equal(
            [stmodels.sum(lag, self.Vx, self.Vt) for lag in self.lags],
            [2.37, 8.76, 12., 12., 9.96, 11.61],
            decimal=2
        )

    def test_default_as_array(self):
        assert_array_almost_equal(
            stmodels.sum(self.lags, self.Vx, self.Vt),
            [2.37, 8.76, 12., 12., 9.96, 11.61],
            decimal=2
        )


class TestProductModel(unittest.TestCase):
    def setUp(self):
        # spatial range = 10;  spatial sill = 5
        self.Vx = lambda h: models.spherical(h, 10, 5)
        # temporal range = 5;  temporal sill = 7
        self.Vt = lambda t: models.cubic(t, 5, 7)

        self.lags = np.array([
            [1.2, 1.0],
            [5.0, 2.5],
            [10., 5.0],
            [12., 7.0],
            [4.2, 7.0],
            [12.0, 3.4]
        ])

    def test_default(self):
        assert_array_almost_equal(
            [stmodels.product(h, self.Vx, self.Vt, 5, 7) for h in self.lags],
            [12.34, 32.37, 35., 35., 35., 35.],
            decimal=2
        )

    def test_default_as_array(self):
        assert_array_almost_equal(
            stmodels.product(self.lags, self.Vx, self.Vt, 5, 7),
            [12.34, 32.37, 35., 35., 35., 35.],
            decimal=2
        )


class TestProductSumModel(unittest.TestCase):
    def setUp(self):
        # spatial range = 10;  spatial sill = 5
        self.Vx = lambda h: models.spherical(h, 10, 5)
        # temporal range = 5;  temporal sill = 7
        self.Vt = lambda t: models.cubic(t, 5, 7)

        self.lags = np.array([
            [1.2, 1.0],
            [5.0, 2.5],
            [10., 5.0],
            [12., 7.0],
            [4.2, 7.0],
            [12.0, 3.4]
        ])

    def test_default(self):
        assert_array_almost_equal(
            [stmodels.product_sum(h, self.Vx, self.Vt, 
                k1=2.2, k2=2.3, k3=4.3, Cx=5, Ct=7) for h in self.lags],
            [35.55, 101.99, 118.6, 118.6, 113.92, 116.91],
            decimal=2
        )

    def test_default_as_array(self):
        assert_array_almost_equal(
            stmodels.product_sum(self.lags, self.Vx, self.Vt, 
                k1=2.2, k2=2.3, k3=4.3, Cx=5, Ct=7),
            [35.55, 101.99, 118.6, 118.6, 113.92, 116.91],
            decimal=2
        )

    def test_with_zero_ks(self):
        assert_array_almost_equal(
            stmodels.product_sum(self.lags, self.Vx, self.Vt, 
                k1=0, k2=0, k3=0, Cx=5, Ct=7),
            [0., 0., 0., 0., 0., 0.],
            decimal=2
        )

    def test_with_all_one(self):
        assert_array_almost_equal(
            stmodels.product_sum(self.lags, self.Vx, self.Vt, 
                k1=1, k2=1, k3=1, Cx=5, Ct=7),
            [14.71, 41.13, 47.  ,47.  ,44.96, 46.61],
            decimal=2
        )

    def test_as_product_model(self):
        assert_array_almost_equal(
            stmodels.product_sum(self.lags, self.Vx, self.Vt, 
                k1=1, k2=0, k3=0, Cx=5, Ct=7),
            stmodels.product(self.lags, self.Vx, self.Vt, 5, 7),
            decimal=2
        )



if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd
import skgstat as skg
from numpy.testing import assert_array_almost_equal


def _get_pan_sample():
    df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'pan_sample.csv'))
    return df[['x', 'y']].values, df.z.values


def test_maxlag_change():
    # get data
    c, v = _get_pan_sample()

    # create a Variogram with default settings
    default = skg.Variogram(c, v)
    maxlag = skg.Variogram(c, v, maxlag=default.bins[-1])

    assert_array_almost_equal(
        default.experimental,
        maxlag.experimental,
        decimal=1
    )



import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage.interpolation import zoom
from scipy.interpolate import griddata

try:
    import plotly.graph_objects as go
except ImportError:
    pass


def matplotlib_plot_2d(stvariogram, kind='contour', ax=None, zoom_factor=100., levels=10, method='fast', **kwargs):
    # get or create the figure
    if ax is not None:
        fig = ax.get_figure()
    else:
        fig, ax = plt.subplots(1, 1, figsize=kwargs.get('figsize', (8, 8)))

    # prepare the meshgrid
    xx, yy = stvariogram.meshbins
    z = stvariogram.experimental.T
    x = xx.flatten()
    y = yy.flatten()

    xxi = zoom(xx, zoom_factor, order=1)
    yyi = zoom(yy, zoom_factor, order=1)

    # interpolation, either fast or precise
    if method.lower() == "fast":
        zi = zoom(z.reshape((stvariogram.t_lags, stvariogram.x_lags)), zoom_factor, order=1, prefilter=False)
    elif method.lower() == "precise":
        # zoom the meshgrid by linear interpolation
        # interpolate the semivariance
        zi = griddata((x, y), z, (xxi, yyi), method='linear')
    else:
        raise ValueError("method has to be one of ['fast', 'precise']")

    # get the bounds
    zmin = np.nanmin(zi)
    zmax = np.nanmax(zi)

    # get the plotting parameters
    lev = np.linspace(0, zmax, levels)
    c = kwargs.get('color', kwargs.get('c', 'k'))
    cmap = kwargs.get('cmap', 'RdYlBu_r')

    # plot
    if kind.lower() == 'contour':
        ax.contour(xxi, yyi, zi, colors=c, levels=lev, vmin=zmin * 1.1, vmax=zmax * 0.9, linewidths=kwargs.get('linewidths', 0.3))
    elif kind.lower() == 'contourf':
        C = ax.contourf(xxi, yyi, zi, cmap=cmap, levels=lev, vmin=zmin *1.1, vmax=zmax * 0.9)
        if kwargs.get('colorbar', True):
            plt.colorbar(C, ax=ax)
    else:
        raise ValueError("%s is not a valid 2D plot" % kind)

    # some labels
    ax.set_xlabel(kwargs.get('xlabel', 'space'))
    ax.set_ylabel(kwargs.get('ylabel', 'time'))
    ax.set_xlim(kwargs.get('xlim', (0, stvariogram.xbins[-1])))
    ax.set_ylim(kwargs.get('ylim', (0, stvariogram.tbins[-1])))

    return fig


def plotly_plot_2d(stvariogram, kind='contour', fig=None, **kwargs):
    # get base data
    x = stvariogram.xbins
    y = stvariogram.tbins
    z = stvariogram.experimental.reshape((len(x), len(y))).T

    # get settings
    showlabels = kwargs.get('showlabels', True)
    colorscale = kwargs.get('colorscale', 'Earth_r')
    smooth = kwargs.get('line_smoothing', 0.0)
    coloring = kwargs.get('coloring', 'heatmap')
    if kind == 'contour':
        coloring = 'lines'
        lw = kwargs.get('line_width', kwargs.get('lw', 2))
        label_color = kwargs.get('label_color', 'black')
    else:
        label_color = kwargs.get('label_color', 'white')
        lw = kwargs.get('line_width', kwargs.get('lw', .3))

    # get the figure
    if fig is None:
        fig = go.Figure()

    # do the plot
    fig.add_trace(
        go.Contour(
            x=x,
            y=y,
            z=z,
            line_smoothing=smooth,
            colorscale=colorscale,
            contours=dict(
                coloring=coloring,
                showlabels=showlabels,
                labelfont=dict(
                    color=label_color,
                    size=kwargs.get('label_size', 14)
                )
            ),
            line_width=lw
        )
    )

    # update the labels
    fig.update_layout(scene=dict(
        xaxis_title=kwargs.get('xlabel', 'space'),
        yaxis_title=kwargs.get('ylabel', 'time')
    ))

    return fig
